/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 10 juillet 2007
   derniere modification : 09/01/2008 */

#ifndef SCENODE_H
#define SCENODE_H

#include <SCE/utils/SCEVector.h>
#include <SCE/utils/SCEMatrix.h>
#include <SCE/utils/SCEList.h>


typedef struct sce_snode SCE_SNode;
typedef void (*SCE_FNodeCallback)(SCE_SNode*, void*);

/* ajoute le 10/07/2007 */
/* revise le 09/01/2008 */
struct sce_snode
{
    SCE_SNode *parent;         /* parent */
    SCE_SList *child;          /* enfants */
    SCE_TMatrix4 matrix;       /* matrice du noeud */
    SCE_TMatrix4 final_matrix; /* matrice finale combinee avec tous les
                                  parents et la matrice du noeud */
    int updated;               /* booleen, true: la matrice finale est a jour */
    void *data;                /* objet relie au noeud */
    SCE_FNodeCallback f;       /* callback appele a chaque Update() */
    void *fparam;              /* second parametre de f */
};


/* cree un nouveau noeud */
SCE_SNode* SCE_Node_Create (void);
/* supprime un noeud */
void SCE_Node_Delete (void*);

/* attache un fils a 'node' */
int SCE_Node_Attach (SCE_SNode*, SCE_SNode*);
/* cree et attache un nouveau fils a 'node' */
SCE_SNode* SCE_Node_AttachNew (SCE_SNode*);
/* detache 'node' de son pere */
void SCE_Node_Detach (SCE_SNode*);
/* detache et supprime 'node' en renvoyant ses donnees */
void* SCE_Node_Erase (SCE_SNode*);

/* insertion d'un noeud entre un noeud et son pere */
int SCE_Node_Insert (SCE_SNode*, SCE_SNode*);

/* multiplie la matrice du noeud par la matrice active d'opengl,
   et defini le resultat comme matrice active */
void SCE_Node_MultMatrix (SCE_SNode*);
/* charge la matrice d'un noeud en tant que matrice active pour opengl */
void SCE_Node_LoadMatrix (SCE_SNode*);

/* renvoie la matrice d'un noeud
   (attention, cela considere que la matrice sera modifiee, toutes les matrices
    finales filles seront donc consideres comme etant non a jour) */
float* SCE_Node_GetMatrix (SCE_SNode*);
/* copie la matrice du noeud dans le pointeur specifie */
void SCE_Node_CopyMatrix (SCE_SNode*, SCE_TMatrix4);

/* renvoie la matrice combinee du noeud avec tous ses parents */
float* SCE_Node_GetFinalMatrix (SCE_SNode*);
/* copie la matrice finale du noeud dans le pointeur specifie */
void SCE_Node_CopyFinalMatrix (SCE_SNode*, SCE_TMatrix4);

/* defini la fonction callback d'un noeud */
void SCE_Node_SetCallback (SCE_SNode*, SCE_FNodeCallback, void*);

/* met a jour la matrice finale d'un noeud */
void SCE_Node_Update (SCE_SNode*);
/* met a jour la matrice finale d'un noeud et de tous ses fils */
void SCE_Node_RecursiveUpdate (SCE_SNode*);

/* indique si le noeud a un papa */
int SCE_Node_HaveParent (SCE_SNode*);
/* renvoie le pere d'un noeud */
SCE_SNode* SCE_Node_GetParent (SCE_SNode*);

/* defini les donnees d'un noeud */
void* SCE_Node_SetData (SCE_SNode*, void*);
/* renvoie les donnees d'un noeud */
void* SCE_Node_GetData (SCE_SNode*);

/* prepare le rendu de l'objet associe au noeud specifie */
void SCE_Node_Use (SCE_SNode*);

#endif /* guard */
