/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 16/05/2007
   derniere modification : 28/06/2007 */


/*
 * Calcule une inertie a une variable
 */

#include <SCE/utils/SCEInert.h>



/* revise le 28/06/2007 */
void SCE_Inert_Init(SCE_SInertVar *ivar)
{
    ivar->var = ivar->real = ivar->toadd = ivar->inert = 0.0f;
    ivar->max_pass = ivar->n_pass = 0;
    ivar->accum = 1;
}


void SCE_Inert_Accum(SCE_SInertVar *ivar, int accum)
{
    ivar->accum = accum;
}


void SCE_Inert_SetParams(SCE_SInertVar *ivar, float inert, int pass)
{
    ivar->inert = inert;
    ivar->max_pass = pass;
}


void SCE_Inert_Set(SCE_SInertVar *ivar, float val)
{
    ivar->toadd += val;
}

float SCE_Inert_Get(SCE_SInertVar *ivar)
{
    return (ivar->var);
}



/* revise le 28/06/2007 */
void SCE_Inert_ComputeVar(SCE_SInertVar *ivar)
{
    /* si aucun nouveau mouvement n'est detecte, on incremente notre varaible
       n_pass. lorsqu'elle aura atteint max_pass, on terminera le mouvement
       (voir le second if plus bas) */
    if(ivar->toadd <= 0.000001f && ivar->toadd >= -0.000001f)
        ivar->n_pass++;
    else
        ivar->n_pass = 0;
    
    /* incrementation de la valeur reelle definitive */
    if(ivar->accum)
        ivar->real += ivar->toadd;
    
    /* si le nombre d'iterations sans mouvement a atteint le maximum autorise,
       on termine notre mouvement */
    if(ivar->n_pass >= ivar->max_pass)
    {
        ivar->n_pass = ivar->max_pass; /* on evite une accumulation */
        if(ivar->accum)
            ivar->var = ivar->real;
        else
            ivar->var += ivar->real + ivar->toadd;
    }
    else
    {
        if(ivar->accum)
            ivar->var += (ivar->real - ivar->var) * (ivar->inert / 100.0f);
        else
            ivar->var = ((ivar->toadd + ivar->real) - ivar->var)
                            * (ivar->inert / 100.0f);
    }
    
    if(!ivar->accum)
        ivar->real += ivar->toadd - ivar->var;
    
    ivar->toadd = 0.0f;
    
    if(ivar->var <= 0.000100f && ivar->var >= -0.000100f)
        ivar->var = 0.0f; /* bicoz division par 100 is not precis a l'infini */
}


void SCE_Inert_Null(SCE_SInertVar *ivar)
{
    if(ivar->accum)
        ivar->var += ivar->toadd;
    else
        ivar->var = ivar->toadd;
    
    ivar->toadd = 0.0f;
}

/*
void SCE_Inert_ComputeVec(SCE_SInertVec *ivec)
{
    
}*/
